# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2017, 2018. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# ******************************************************************

import numpy as np
import sys
import math
from numba import cuda

from _snap_ml.utils import _is_mpi_enabled, _is_data_snapml_partition_type, _convert_labs
from _snap_ml.utils import _verify_compat_obj_dtype, _verify_inference_dtype
_mpi_enabled = _is_mpi_enabled()

if sys.version_info[0] < 3:
    if _mpi_enabled:
        import libsnapmlmpi2 as libsnapml
    else:
        import libsnapmllocal2 as libsnapml
else:
    if _mpi_enabled:
        import _snap_ml.libsnapmlmpi3 as libsnapml
    else:
        import _snap_ml.libsnapmllocal3 as libsnapml


class LogisticRegression:

    """
    Logistic Regression classifier

    This class implements regularized logistic regression using the IBM Snap ML solver.
    It supports both local and distributed(MPI) methods of the Snap ML solver. It
    can be used for both binary and multi-class classification problems. For multi-class
    classification it predicts only classes (no probabilities). It handles both dense and 
    sparse matrix inputs. Use csr, csc, ndarray, deviceNDArray or SnapML data partition format
    for training and csr, ndarray or SnapML data partition format for prediction.
    DeviceNDArray input data format is currently not supported for training with MPI implementation.
    We recommend the user to first normalize the input values.

    Parameters
    ----------
    max_iter : int, default : 1000
        Maximum number of iterations used by the solver to converge.

    regularizer : float, default : 1.0
        Regularization strength. It must be a positive float. 
        Larger regularization values imply stronger regularization.
    
    use_gpu : bool, default : False
        Flag for indicating the hardware platform used for training. If True, the training 
        is performed using the GPU. If False, the training is performed using the CPU.

    device_ids : array-like of int, default : []
        If use_gpu is True, it indicates the IDs of the GPUs used for training.
        For single-GPU training, set device_ids to the GPU ID to be used for training, e.g., [0].
        For multi-GPU training, set device_ids to a list of GPU IDs to be used for training, e.g., [0, 1].

    class_weight : 'balanced' or None, optional
        If set to 'None', all classes will have weight 1.

    dual : bool, default : True
        Dual or primal formulation. 
        Recommendation: if n_samples > n_features use dual=True.

    verbose : bool, default : False
        If True, it prints the training cost, one per iteration. Warning: this will increase the
        training time. For performance evaluation, use verbose=False.

    num_threads : int, default : 1 
        The number of threads used for running the training. The value of this parameter
        should be a multiple of 32 if the training is performed on GPU (use_gpu=True).

    penalty : str, default : "l2"
        The regularization / penalty type. Possible values are "l2" for L2 regularization (LogisticRegression)
        or "l1" for L1 regularization (SparseLogisticRegression). L1 regularization is possible only for the primal
        optimization problem (dual=False).

    tol : float, default : 0.001
        The tolerance parameter. Training will finish when maximum change in model coefficients is less than tol. 

    return_training_history : str or None, default : None
        How much information about the training should be collected and returned by the fit function. By 
        default no information is returned (None), but this parameter can be set to "summary", to obtain 
        summary statistics at the end of training, or "full" to obtain a complete set of statistics 
        for the entire training procedure. Note, enabling either option will result in slower training.
        return_training_history is not supported for DeviceNDArray input format. 

    privacy : bool, default : False
        Train the model using a differentially private algorithm.
        Currently not supported for MPI implementation.

    eta : float, default : 0.3
        Learning rate for the differentially private training algorithm.
        Currently not supported for MPI implementation.

    batch_size : int, default : 100
        Mini-batch size for the differentially private training algorithm.
        Currently not supported for MPI implementation.

    privacy_epsilon : float, default : 10.0
        Target privacy gaurantee. Learned model will be (privacy_epsilon, 0.01)-private.
        Currently not supported for MPI implementation.

    grad_clip : float, default: 1.0
        Gradient clipping parameter for the differentially private training algorithm.
        Currently not supported for MPI implementation.

    fit_intercept : bool, default : False
        Add bias term -- note, may affect speed of convergence, especially for sparse datasets.

    intercept_scaling : float, default : 1.0
        Scaling of bias term. The inclusion of a bias term is implemented by appending an additional feature to the 
        dataset. This feature has a constant value, that can be set using this parameter.

    Attributes
    ----------
    coef_ : array-like, shape (n_features, 1) for binary classification or
        (n_features, n_classes) for multi-class classification.
        Coefficients of the features in the trained model.

    support_ : array-like
        Indices of the features that contribute to the decision. (only available for L1)
        Currently not supported for MPI implementation.

    model_sparsity_ : float
        fraction of non-zeros in the model parameters. (only available for L1)
        Currently not supported for MPI implementation.
    """

    def __init__(self,  max_iter=1000, 
                        regularizer=1.0, 
                        device_ids=[], 
                        verbose=False, 
                        use_gpu=False, 
                        class_weight=None, 
                        dual=True, 
                        num_threads=1, 
                        penalty="l2",
                        tol=0.001,
                        return_training_history=None,
                        privacy=False,
                        eta=0.3,
                        batch_size=100,
                        privacy_epsilon=10,
                        grad_clip=1,
                        fit_intercept=False,
                        intercept_scaling=1.0):

        self.max_iter_ = max_iter
        if max_iter < 1:
            raise ValueError("Parameter max_iter: value too small. Should be an integer >= 1.")	

        self.regularizer_ = regularizer
        if regularizer < 0.0001:
            print("[Warning] Parameter regularizer: value might be too small.")
        elif regularizer <= 0:
            raise ValueError("Parameter regularizer: value should be positive.")

        if any(val < 0 for val in device_ids):
            raise ValueError("Parameter device_id: value too small. Should be an integer >= 0.")
        device_ids.sort()
        self.device_ids_ = np.array(device_ids).astype(np.uint32)

        self.verbose_ = verbose
        if not isinstance(verbose, (bool)):
            raise TypeError("Parameter verbose: parameter type not valid. Should be a bool {True, False}.")

        self.class_weight_ = class_weight
        if (not class_weight == 'balanced') and (not class_weight == None):
            raise ValueError("Parameter class_weight: parameter value not valid. Should be {'balanced', 'None'}.")

        self.use_gpu_ = use_gpu
        if not isinstance(use_gpu, (bool)):
            raise TypeError("Parameter use_gpu: parameter type not valid. Should be a bool {True, False}.")

        if (use_gpu == True):
            if (not num_threads % 32 == 0):
                print("[Info] If set num_threads should be a multiple of 32. GPU training will run with num_threads=256.")
                self.num_threads_ = 256
            else:
                self.num_threads_ = num_threads
        else:
            if (num_threads > 1):
                print("[Info] Training will run in multi-threaded mode on CPU.")
                self.num_threads_ = num_threads
            else:
                print("[Info] Training will run in single-threaded mode on CPU.")
                self.num_threads_ = 1

        if (self.use_gpu_ == False) and (len(device_ids) > 0):
            print("[Warning] Parameter device_ids will be ignored. use_gpu was set to False. The training will run on the CPU.")

        if (self.use_gpu_ == True) and (len(device_ids) == 0):
            print("[Warning] Parameter device_ids not set. The training will run on GPU 0.")

        self.dual_ = dual
        if not isinstance(dual, (bool)):
            raise TypeError("Parameter dual: parameter type not valid. Should be a bool {True, False}.")
 
        if not _mpi_enabled:
            self.uniq_labs_ = []

        self.penalty_ = penalty
        if ((self.penalty_ != "l1") and (self.penalty_ != "l2")):
            raise ValueError("Parameter penalty: wrong value. Should be l1 or l2.")

        if ((self.penalty_ == "l1") and (self.dual_)):
            raise ValueError("Sparse logistic regression with L1 regularization is supported only for primal optimization problems. Set dual to False.")

        self.tol_ = tol
        if(self.tol_ < 0):
            raise ValueError("Parameter tol: must be greater than or equal to 0")

        self.return_training_history_ = return_training_history
        if(self.return_training_history_ not in [None, "summary", "full"]):
            raise ValueError("Parameter return_training_history must be None, \'summary\' or \'full\'")

        self.fit_intercept_ = fit_intercept
        if(self.fit_intercept_ not in [False, True]):
            raise ValueError("Parameter fit_intercept must be either False or True")

        self.intercept_scaling_ = intercept_scaling

        self.privacy_ = privacy
        if not isinstance(privacy, (bool)):
            raise TypeError("Parameter privacy: parameter type not valid. Should be a bool {True, False}.")

        if(_mpi_enabled and self.privacy_):
            raise ValueError("Privacy not supported for MPI execution. Set privacy=False.")

        self.eta_ = eta
        if(self.eta_ <= 0):
            raise ValueError("Parameter eta must be greater than 0")
    
        self.batch_size_ = batch_size
        if(self.batch_size_ < 0):
            raise ValueError("Parameter batch_size must be greater than 0")
    
        self.grad_clip_ = grad_clip
        if(self.grad_clip_ < 0):
            raise ValueError("Parameter grad_clip must be greater or equal to 0")
    
        self.privacy_epsilon_ = privacy_epsilon
        if(self.privacy_epsilon_ <= 0):
            raise ValueError("Parameter privacy_epsilon must be greater than 0")
        
        if(self.privacy_ and self.dual_):
            raise ValueError("Privacy only supported for primal objective functions.")
        
        if(self.privacy_ and self.penalty_ == "l1"):
            raise ValueError("Privacy only supported for L2-regularized objective functions.")
    
        if(self.privacy_ and self.use_gpu_):
            print("[Warning] Privacy not supported by GPU solvers. Will train on CPU")

        if(self.privacy_ and self.fit_intercept_):
            raise ValueError("Privacy not supported with fit_intercept=True.")

        self.labels_flag = False

    def get_params(self):

        """
        Get the values of the model parameters.

        Returns
        -------
        params : dict
        """

        params = {}
        params['max_iter'] = self.max_iter_
        params['regularizer'] = self.regularizer_
        params['use_gpu'] = self.use_gpu_
        params['device_ids'] = self.device_ids_
        params['class_weight'] = self.class_weight_
        params['dual'] = self.dual_
        params['verbose'] = self.verbose_
        params['num_threads'] = self.num_threads_
        params['penalty'] = self.penalty_
        params['tol'] = self.tol_
        params['return_training_history'] = self.return_training_history_
        params['privacy'] = self.privacy_
        params['eta'] = self.eta_
        params['batch_size'] = self.batch_size_
        params['grad_clip'] = self.grad_clip_
        params['privacy_epsilon'] = self.privacy_epsilon_
        params['fit_intercept'] = self.fit_intercept_
        params['intercept_scaling'] = self.intercept_scaling_

        return params

    def fit(self, X_train, y_train=None):

        """
        Fit the model according to the given train data.

        Parameters
        ----------
        X_train : Train dataset. Supports the following input data-types :
            1. Sparse matrix (csr_matrix, csc_matrix) or dense matrix (ndarray)
            2. DeviceNDArray. Not supported for MPI execution.
            3. SnapML data partition of type DensePartition, SparsePartition or ConstantValueSparsePartition

        y_train : The target corresponding to X_train.
            If X_train is sparse matrix or dense matrix, y_train should be array-like of shape = (n_samples,)
            In case of deviceNDArray, y_train should be array-like of shape = (n_samples, 1)
            For binary classification the labels should be {-1, 1} or {0, 1}.
            If X_train is SnapML data partition type, then y_train is not required (i.e. None).

        Returns
        -------
        self : object
        """

        # Check if the balanced parameter was set
        if (self.class_weight_ == 'balanced'):
            balanced = True
        else:
            balanced = False

        X_train_ptr_ = ""
        X_train_type_ = -1
        is_regression = 0
       
        gpu_data_ptr = 0
        gpu_lab_ptr = 0
        num_ex = 0
        num_ft = 0
        num_nz = 0
        indptr = np.array([])
        indices = np.array([])
        data = np.array([])
        labs = np.array([])
        self.num_train_unique_labs = 0
        self.model_size = 0
        # by default gpu_matrix is set to False.
        # It will be set to True if xtrain and ytrain is coming from GPU device.
        gpu_matrix = False

        # DeviceNDArray support related checks  
        if (type(X_train).__name__ == "DeviceNDArray" and type(y_train).__name__ == "DeviceNDArray"):
            if (X_train.dtype != "float32" or y_train.dtype != "float32"):
                raise TypeError("X_train and y_train should be in float32")
            if(len(self.device_ids_)>1):
                raise ValueError("MultiGPU support is not there currently for cuDF")
            if(len(self.device_ids_)==1 and self.device_ids_[0]!=0):
                raise ValueError("cuDF only runs on GPU id 0.")
            if _mpi_enabled:
                raise TypeError("DeviceNDArray is not supported for MPI")
            if (self.use_gpu_ == False):
                raise ValueError("To use DeviceNDArray, use_gpu true is mandatory option")
            if(self.return_training_history_ is not None):
                raise ValueError("training_history is not supported for DeviceNDArray")


        # Ensure y_train is passed in, if data is not snapml partitioned format
        if (not _is_data_snapml_partition_type(X_train)) and (y_train is None):
            raise ValueError("Parameter y_train: Input data is required.")

        if _is_data_snapml_partition_type(X_train):
            X_train_ptr_ = X_train.ptr_
            X_train_type_ = X_train.type_
            self.privacy_ = False
            self.eta_ = 0
            self.batch_size_ = 0
            self.grad_clip_ = 0
            self.privacy_epsilon_ = 0
        else:
            if _mpi_enabled:
                if(type(X_train).__name__ == "ndarray"):
                    num_ex = len(y_train)
                    if (num_ex == X_train.shape[0]):
                        #Dual formation
                        num_ft = X_train.shape[1]
                    else:
                        #Primal formation
                        num_ft = X_train.shape[0]
                elif(type(X_train).__name__ == "csr_matrix" or type(X_train).__name__ == "csc_matrix"):
                    _verify_compat_obj_dtype(self.dual_, type(X_train).__name__)
                    num_ex = X_train.shape[0]
                    num_ft = X_train.shape[1]
            else:
                num_ex = len(y_train)

                # Check train dataset format
                if(self.dual_ == False):
                    if(self.privacy_ == True):                
                        if (X_train.shape[0] != num_ex):
                            raise ValueError("X_train.shape[0] must be equal to number of examples.")
                        
                        if type(X_train).__name__ not in ["csr_matrix", "ndarray", "DeviceNDArray"]:
                            raise ValueError("Private solver requires training data in csr_matrix or ndarray or DeviceNDArray")                    
                        num_ft = X_train.shape[1]                
                    else:            
                        if (type(X_train).__name__ == "csr_matrix"):
                            X_train = X_train.tocsc()
                            num_ft = X_train.shape[1]
                        elif(type(X_train).__name__ == "csc_matrix"):
                            num_ft = X_train.shape[1]
                        elif(type(X_train).__name__ == "ndarray" or type(X_train).__name__ == "DeviceNDArray"):
                            if (X_train.shape[0] == X_train.shape[1]):
                                print("[Warning] Training will run using the dual formulation. Each row in the input train dataset is assumed to be a training sample.")
                                self.dual_ = True
                                num_ft = X_train.shape[0]
                            elif(X_train.shape[0] == num_ex):
                                if (type(X_train).__name__ == "DeviceNDArray"):
                                    X_train = X_train.transpose()
                                    if (X_train.is_c_contiguous() == False):
                                        raise ValueError("X_train is not c_contiguous after transpose") 
                                else:
                                    X_train = X_train.transpose().copy()
                                num_ft = X_train.shape[0]
                            else:
                                num_ft = X_train.shape[0]
                else:
                    if (type(X_train).__name__ == "csr_matrix"):
                        num_ft = X_train.shape[1]
                    elif(type(X_train).__name__ == "csc_matrix"):
                        X_train = X_train.tocsr()
                        num_ft = X_train.shape[1]
                    elif(type(X_train).__name__ == "ndarray" or type(X_train).__name__ == "DeviceNDArray"):
                        if (X_train.shape[0] == X_train.shape[1]):
                            print("[Warning] Training will run using the dual formulation. Each row in the input train dataset is assumed to be a training sample.")
                            num_ft = X_train.shape[0]
                        elif(X_train.shape[0] == num_ex):
                            num_ft = X_train.shape[1]
                        else:
                            X_train = X_train.transpose().copy()
                            num_ft = X_train.shape[1]

            if (type(X_train).__name__ == "csr_matrix" or type(X_train).__name__ == "csc_matrix"):
                num_nz    = X_train.nnz
                indptr    = X_train.indptr.astype(np.uint64)
                data      = X_train.data.astype(np.float32)
                indices   = X_train.indices.astype(np.uint32)
            elif (type(X_train).__name__ == "ndarray"):
                num_nz    = num_ex * num_ft
                data      = X_train.astype(np.float32)
                indptr    = np.array([])
                indices   = np.array([])
            elif (type(X_train).__name__ == "DeviceNDArray"):
                num_nz    = num_ex * num_ft
                data      = np.array([]).astype(np.float32)
                indptr    = np.array([])
                indices   = np.array([])
                gpu_matrix = True
                gpu_data_ptr   = X_train.device_ctypes_pointer.value
                gpu_lab_ptr    = y_train.device_ctypes_pointer.value
                # To find unique lebel we are creating copy lab into host as ndarry ; So that np.unique() can be performed
                # TODO : avoid the separate host copy
                y_train_host   = y_train.copy_to_host()
                y_train_device = y_train
                y_train        = y_train_host
                labs           = y_train.astype(np.float32)
            else:
                raise TypeError("X_train should be in csr_matrix, csc_matrix, ndarray or DeviceNDArray format.")
  

            # Extract set of unique labels from y_train
            self.uniq_labs_ = np.unique(y_train)

        if(self.return_training_history_ is None):
            return_training_code = 0
        elif(self.return_training_history_ is 'summary'):
            return_training_code = 1
        else:
            return_training_code = 2

        # Set parameter transpose
        if (self.dual_ == False):
            if(self.privacy_ == True):
                transpose = False
            else:
                transpose = True
        else: 
            transpose = False

        # Check and modify the labels of y_train
        _multiclass = False
        if ((y_train is not None) and (not _is_data_snapml_partition_type(y_train))):
            if _mpi_enabled:
                labs = y_train.astype(np.float32)
            else: #local
                if(len(self.uniq_labs_) == 1):
                    raise ValueError("There must be at least two unique label values in the training data")
                elif (len(self.uniq_labs_) == 2):
                    _multiclass = False
                    if ( (self.uniq_labs_[0] == 0 and self.uniq_labs_[1] == 1)):
                        self.labels_flag = True
                        if (gpu_matrix == False ):
                            labs = 2.0 * y_train.astype(np.float32) - 1.0
                        else:
                            labs = 2.0 * y_train.astype(np.float32) - 1.0
                            index = 16
                            th_per_blk = (index,index)
                            bl_per_grd_x = math.ceil(y_train_device.shape[0] / th_per_blk[0])
                            bl_per_grd_y = math.ceil(y_train_device.shape[1] / th_per_blk[1])
                            bl_per_grd  = (bl_per_grd_x,bl_per_grd_y)
                            _convert_labs[bl_per_grd, th_per_blk](y_train_device)
                    elif (not(self.uniq_labs_[0] == -1 and self.uniq_labs_[1] == 1)):
                        raise ValueError("The labels in the train dataset should be {-1, 1} or {0, 1}.")
                    else:
                        if (gpu_matrix == False):
                            labs = y_train.astype(np.float32)
                else:
                    labs = y_train.astype(np.float32)
                    _multiclass = True

        if (type(X_train).__name__ == "DeviceNDArray" and type(y_train).__name__ == "DeviceNDArray"):
            if ( len(self.uniq_labs_) > 2 ):
                raise ValueError("Multiclass is not supported for gpu matrix")

        # Train the model
        out_fit = libsnapml.lr_fit( self.max_iter_,
                                    self.regularizer_,
                                    1,
                                    self.verbose_,
                                    balanced,
                                    self.use_gpu_,
                                    self.num_threads_,
                                    transpose,
                                    num_ex,
                                    num_ft,
                                    num_nz,
                                    indptr,
                                    indices,
                                    data,
                                    labs,
                                    gpu_data_ptr,
                                    gpu_lab_ptr,
                                    gpu_matrix,
                                    self.device_ids_,
                                    X_train_ptr_,
                                    X_train_type_,
                                    self.penalty_,
                                    self.tol_,
                                    return_training_code,
                                    self.privacy_,
                                    self.eta_,
                                    self.batch_size_,
                                    self.grad_clip_,
                                    self.privacy_epsilon_,
                                    0, #is_regression
                                    self.fit_intercept_,
                                    self.intercept_scaling_)

        self.model_ = out_fit[0]
        self.training_history_ = out_fit[1]
        if not _mpi_enabled:
            self.support_ = np.array(out_fit[2], np.int32)
            #self.model_sparsity_ = self.support_.shape[0]/num_ft
            self.model_sparsity_ = self.support_.shape[0]/len(out_fit[0])
            self.n_iter_ = out_fit[3]

        # training_metadata
        training_metadata = out_fit[4]
        self.model_size = training_metadata['model_size']
        self.num_train_unique_labs = int(training_metadata['num_uniq_labs'])
        self.uniq_labs_ = np.array(training_metadata['uniq_labs'])
        if _is_data_snapml_partition_type(X_train) or _mpi_enabled:
            #For snap-ml-local with non-snap-partition-type 'labels_flag' is set differently
            self.labels_flag = int(training_metadata['labs_converted'])

        self.coef_ = self.model_.T
        # set labels to original values - only in the case of multi-class classification
        # libsnapmlmpi.lr_restore_dataset(data.ptr_);

        return self.training_history_

    # The test dataset should be in csr_matrix format
    def predict_proba(self, X, num_threads=0):

        """
        Probability estimates

        The returned probability estimates for the two classes.
        Only for binary classification. 

        Parameters
        ----------
        X : Dataset used for predicting probability estimates. Supports the following input data-types :
            1. Sparse matrix (csr_matrix, csc_matrix) or dense matrix (ndarray)
            2. SnapML data partition of type DensePartition, SparsePartition or ConstantValueSparsePartition

        num_threads : int, default : 0
            Number of threads used to run inference.
            By default inference runs with maximum number of available threads.

        Returns
        -------
        proba: Returns the probability of the sample of each of the two classes for local implementation :
                    array-like of shape = (n_samples, 2)
               Returns the probability of the sample to be a positive example for MPI :
                    array-like of shape = (n_samples,)
        """
        if (self.num_train_unique_labs > 2):
            raise TypeError("Only predict support in multi-class classification mode. User input: predict_proba.")

        is_regression = 0
        num_ex = 0
        num_ft = 0
        indptr = np.array([])
        indices =  np.array([])
        data = np.array([])
        X_ptr_ = ""
        X_type_ = -1
        num_train_unique_labs = 0
        if _is_data_snapml_partition_type(X):
            X_ptr_ = X.ptr_
            X_type_ = X.type_
            num_train_unique_labs = self.num_train_unique_labs
        else:
            if _mpi_enabled:
                _verify_inference_dtype(type(X).__name__)

            num_ex  = X.shape[0]
            num_ft  = X.shape[1]
            num_train_unique_labs = len(self.uniq_labs_)

            if (type(X).__name__ == "csc_matrix"):
                X = X.tocsr()
                indptr    = X.indptr.astype(np.uint64)
                data      = X.data.astype(np.float32)
                indices   = X.indices.astype(np.uint32)
            elif (type(X).__name__ == "csr_matrix"):
                indptr    = X.indptr.astype(np.uint64)
                data      = X.data.astype(np.float32)
                indices   = X.indices.astype(np.uint32)
            elif (type(X).__name__ == "ndarray"):
                data      = X.astype(np.float32)
                indptr    = np.array([])
                indices   = np.array([])
            else:
                raise ValueError("X should be in csc_matrix, csr_matrix or ndarray format.")

            if ( (self.fit_intercept_ == True) and (type(X).__name__ == "csr_matrix" or type(X).__name__ == "csc_matrix") ):
                print("[Warning] May affect speed of convergence for sparse datasets as fit_intercept is set.")

            if (len(self.uniq_labs_) > 2):
                raise ValueError("predict_proba is currently not supported for multi-class classification. Use predict to compute the class estimates.")

        # Compute probabilities
        proba, proba_metadata = libsnapml.lr_predict_proba(num_ex,
                                           num_ft,
                                           indptr,
                                           indices,
                                           data,
                                           self.model_,
                                           X_ptr_,
                                           X_type_,
                                           self.model_size,
                                           num_train_unique_labs,
                                           is_regression,
                                           num_threads,
                                           self.fit_intercept_,
                                           self.intercept_scaling_)

        # First element is the number of examples in X_test
        return np.array(proba)

    def predict(self, X, num_threads=0):

        """
        Class predictions

        The returned class estimates.

        Parameters
        ----------
        X : Dataset used for predicting class estimates. Supports the following input data-types :
            1. Sparse matrix (csr_matrix, csc_matrix) or dense matrix (ndarray)
            2. SnapML data partition of type DensePartition, SparsePartition or ConstantValueSparsePartition

        num_threads : int, default : 0
            Number of threads used to run inference.
            By default inference runs with maximum number of available threads.

        Returns
        -------
        proba: array-like, shape = (n_samples,)
            Returns the predicted class of the sample.
        """

        is_regression = 0
        num_ex = 0
        num_ft = 0
        indptr = np.array([])
        indices =  np.array([])
        data = np.array([])
        X_ptr_ = ""
        X_type_ = -1
        num_train_unique_labs = 0
        if _is_data_snapml_partition_type(X):
            X_ptr_ = X.ptr_
            X_type_ = X.type_
            num_train_unique_labs = self.num_train_unique_labs
        else:
            if _mpi_enabled:
                _verify_inference_dtype(type(X).__name__)

            num_ex  = X.shape[0]
            num_ft  = X.shape[1]
            num_train_unique_labs = len(self.uniq_labs_)
        
            if (type(X).__name__ == "csc_matrix"):
                X = X.tocsr()
                indptr    = X.indptr.astype(np.uint64)
                data      = X.data.astype(np.float32)
                indices   = X.indices.astype(np.uint32)
            elif (type(X).__name__ == "csr_matrix"):
                indptr    = X.indptr.astype(np.uint64)
                data      = X.data.astype(np.float32)
                indices   = X.indices.astype(np.uint32)
            elif (type(X).__name__ == "ndarray"):
                data      = X.astype(np.float32)
                indptr    = np.array([])
                indices   = np.array([])
            else:
                raise ValueError("X should be in csc_matrix, csr_matrix or ndarray format.")

        pred = []

        # Generate predictions
        pred, pred_metadata  = libsnapml.lr_predict( num_ex,
                                                     num_ft,
                                                     indptr,
                                                     indices,
                                                     data,
                                                     self.model_,
                                                     X_ptr_,
                                                     X_type_,
                                                     self.model_size,
                                                     num_train_unique_labs,
                                                     is_regression,
                                                     num_threads,
                                                     self.fit_intercept_,
                                                     self.intercept_scaling_)

        # For binary classification transform the labels if needed
        if ((num_train_unique_labs <= 2) and (self.labels_flag)):
            pred = (pred + 1) / 2

        # Multi-class classification case
        if (num_train_unique_labs > 2):
            pred = np.argmax(pred, axis=0)

            class_to_index_mapping = np.array([])
            for class_ in self.uniq_labs_:
                class_to_index_mapping = np.append(class_to_index_mapping, class_)

            for example_index in range(len(pred)):
                pred[example_index] = class_to_index_mapping[pred[example_index]]

        return pred

    # The test dataset should be in csr_matrix format
    def predict_log_proba(self, X, num_threads=0):

        """
        Log of probability estimates

        The returned log-probability estimates for the two classes.
        Only for binary classification.

        Parameters
        ----------
        X : Dataset used for predicting log-probability estimates. Supports the following input data-types :
            1. Sparse matrix (csr_matrix, csc_matrix) or dense matrix (ndarray)
            2. SnapML data partition of type DensePartition, SparsePartition or ConstantValueSparsePartition

        num_threads : int, default : 0
            Number of threads used to run inference.
            By default inference runs with maximum number of available threads.

        Returns
        -------
        proba: Returns the log-probability of the sample of each of the two classes for local implementation :
                    array-like of shape = (n_samples, 2)
               Returns the log-probability of the sample to be a positive example for MPI :
                    array-like of shape = (n_samples,)
        """

        is_regression =0
        num_ex = 0
        num_ft = 0
        indptr = np.array([])
        indices =  np.array([])
        data = np.array([])
        X_ptr_ = ""
        X_type_ = -1
        num_train_unique_labs = 0

        if _is_data_snapml_partition_type(X):
            X_ptr_ = X.ptr_
            X_type_ = X.type_
            num_train_unique_labs = self.num_train_unique_labs
        else:
            if _mpi_enabled:
                _verify_inference_dtype(type(X).__name__)

            num_ex  = X.shape[0]
            num_ft  = X.shape[1]
            num_train_unique_labs = len(self.uniq_labs_)

            if (type(X).__name__ == "csc_matrix"):
                X = X.tocsr()
                indptr    = X.indptr.astype(np.uint64)
                data      = X.data.astype(np.float32)
                indices   = X.indices.astype(np.uint32)
            elif (type(X).__name__ == "csr_matrix"):
                indptr    = X.indptr.astype(np.uint64)
                data      = X.data.astype(np.float32)
                indices   = X.indices.astype(np.uint32)
            elif (type(X).__name__ == "ndarray"):
                data      = X.astype(np.float32)
                indptr    = np.array([])
                indices   = np.array([])
            else:
                raise ValueError("X should be in csc_matrix, csr_matrix or ndarray format.")

        if (num_train_unique_labs > 2):
            raise ValueError("predict_proba is currently not supported for multi-class classification. Use predict to compute the class estimates.")

        # Compute log probabilities
        proba, proba_metadata = libsnapml.lr_predict_log_proba(num_ex,
                                                               num_ft,
                                                               indptr,
                                                               indices,
                                                               data,
                                                               self.model_,
                                                               X_ptr_,
                                                               X_type_,
                                                               self.model_size,
                                                               num_train_unique_labs,
                                                               is_regression,
                                                               num_threads,
                                                               self.fit_intercept_,
                                                               self.intercept_scaling_)

        return np.array(proba)

