# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2018, 2019. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

from __future__ import print_function
import os
import time
import argparse
from scipy.sparse import load_npz
import numpy as np

# pai4sk explicit imports for snap_ml APIs
from pai4sk.sml_io import load_from_snap_format
from pai4sk import LogisticRegression as snapmlLR
from pai4sk.sml_metrics import log_loss as snapLOGLOSS
from pai4sk.sml_metrics import accuracy_score as snapACC

# pai4sk explicit imports for sklearn APIs
from pai4sk.linear_model import LogisticRegressionSklearn as sklearnLR
from pai4sk.metrics import log_lossSklearn, accuracy_scoreSklearn

defaultPath = os.getcwd()

CLI=argparse.ArgumentParser()
CLI.add_argument(
   "--data_path",
   type=str,
   default=defaultPath
)
CLI.add_argument(
   "--device_ids",
   nargs="*",
   type=int,
   default=[]
)
CLI.add_argument(
    "--use_gpu",
    action = 'store_true',
    default=False
)
CLI.add_argument(
    "--skip_sklearn",
    action = 'store_true',
    default=False
)
CLI.add_argument(
   "--use_balanced_class_weights",
   action = 'store_true',
   default=False
)
args = CLI.parse_args()

print("[pai4sk] Data loading...")
t0 = time.time()
# Datasets paths (snap.ml distributed data loading)
train_filename_snap = args.data_path + "/data/criteo.kaggle2014.train.snap"
test_filename_snap = args.data_path + "/data/criteo.kaggle2014.test.snap"

## snapML
train_data = load_from_snap_format(train_filename_snap)
test_data = load_from_snap_format(test_filename_snap)

load_time_snap = time.time() - t0

# Create snapML Logistic Regression
if args.use_gpu:
  num_threads = 256
  cpu_gpu = "GPU"
else:
  num_threads = 8
  cpu_gpu = "CPU"


if args.use_balanced_class_weights:
  class_weight = "balanced"
else:
  class_weight = None

snapml_lr = snapmlLR(dual=True, use_gpu=args.use_gpu, device_ids=args.device_ids,
                     num_threads=num_threads, class_weight=class_weight)

print("[pai4sk] Training model on %s..." % (cpu_gpu))
# Fit the model and time it
snapml_t0 = time.time()
snapml_lr.fit(train_data)
snapml_time = time.time() - snapml_t0

print("[pai4sk] Running prediction...")
# Perform inference on test data
pred = snapml_lr.predict(test_data)
proba = snapml_lr.predict_proba(test_data)


# Compute accuracy
snapml_accuracy = snapACC(test_data, pred)
snapml_logloss = snapLOGLOSS(test_data, proba)

# Print off SnapML result
print("[pai4sk] [Data loading] Time = %.4f" % (load_time_snap))
print("[pai4sk] [%s Training] Accuracy = %.4f" % (cpu_gpu, snapml_accuracy))
print("[pai4sk] [%s Training] LogLoss = %.4f" % (cpu_gpu, snapml_logloss))
print("[pai4sk] [%s Training] Training time = %.4f" % (cpu_gpu, snapml_time))

## SKLEARN
if args.skip_sklearn:
    print('Execution using scikit-learn is skipped.')
    exit()

print("[sklearn] Data loading...")

t0 = time.time()
# Datasets paths (sklearn loading)
x_train_filename_sklearn = args.data_path + "/data/criteo.kaggle2014.X_train.npz"
y_train_filename_sklearn = args.data_path + "/data/criteo.kaggle2014.y_train.npy"
x_test_filename_sklearn = args.data_path + "/data/criteo.kaggle2014.X_test.npz"
y_test_filename_sklearn = args.data_path + "/data/criteo.kaggle2014.y_test.npy"

X_train = load_npz(x_train_filename_sklearn)
X_test  = load_npz(x_test_filename_sklearn)
y_train = np.load(y_train_filename_sklearn)
y_test  = np.load(y_test_filename_sklearn)

load_time_sklearn = time.time() - t0

# Import model
sklearn_lr = sklearnLR(dual=True, tol=0.001)

print("[sklearn] Training model...")
# Fit the model and time it
sklearn_t0 = time.time()
sklearn_lr_model = sklearn_lr.fit(X_train, y_train)
sklearn_time = time.time() - sklearn_t0

print("[sklearn] Running prediction...")
# Perform inference on test data
pred_proba = sklearn_lr.predict_proba(X_test)
pred = sklearn_lr.predict(X_test)

from sklearn.metrics import log_loss, accuracy_score
sklearn_accuracy = accuracy_scoreSklearn(y_test, pred) 
sklearn_logloss = log_lossSklearn(y_test, pred_proba)

print("[sklearn] [Data loading] Time = %.4f" % (load_time_sklearn))
print("[sklearn] Accuracy = %.4f" % (sklearn_accuracy))
print("[sklearn] LogLoss = %.4f" % (sklearn_logloss))
print("[sklearn] Training time = %.4f" % (sklearn_time))
